use std::cell::RefCell;
use std::rc::Rc;
use std::sync::{Arc, RwLock};
use std::thread;
use std::time::Instant;

use im::HashMap;
use libpulse_binding::callbacks::ListResult;
use libpulse_binding::context::introspect::{Introspector, SinkInfo, SinkInputInfo, SourceInfo};
use libpulse_binding::context::subscribe::{Facility, InterestMaskSet};
use libpulse_binding::context::{subscribe, Context, FlagSet};
use libpulse_binding::def::Retval;
use libpulse_binding::mainloop::standard::{IterateResult, Mainloop};
use libpulse_binding::operation::Operation;
use libpulse_binding::volume::{ChannelVolumes, Volume};
use tokio::sync::broadcast;
use tokio::sync::broadcast::Receiver;

pub type PaEntityId = u32;

#[derive(Debug, Eq, PartialEq, Copy, Clone)]
pub enum PaEntityKind {
    Source,
    Sink,
    SinkInput,
}

#[derive(Debug, Eq, PartialEq, Clone)]
pub enum PaEntityMetadata {
    Source {
        name: String,
        description: String,
    },
    Sink {
        name: String,
        description: String,
    },
    SinkInput {
        description: String,
        binary_name: Option<String>,
        application_name: Option<String>,
    },
}

#[derive(Debug, Clone, PartialEq)]
pub struct PaEntityState {
    id: PaEntityId,
    channel_volumes: ChannelVolumes,
    is_muted: bool,
    metadata: PaEntityMetadata,
}

impl PaEntityState {
    pub fn id(&self) -> &PaEntityId {
        &self.id
    }

    pub fn kind(&self) -> PaEntityKind {
        match &self.metadata {
            PaEntityMetadata::Source { .. } => PaEntityKind::Source,
            PaEntityMetadata::Sink { .. } => PaEntityKind::Sink,
            PaEntityMetadata::SinkInput { .. } => PaEntityKind::SinkInput,
        }
    }

    pub fn metadata(&self) -> &PaEntityMetadata {
        &self.metadata
    }

    pub fn channel_volumes(&self) -> Vec<f32> {
        self.channel_volumes.get().iter().map(|v| v.0 as f32 / Volume::NORMAL.0 as f32).collect()
    }

    pub fn is_muted(&self) -> bool {
        self.is_muted
    }
}

impl From<&SourceInfo<'_>> for PaEntityState {
    fn from(value: &SourceInfo) -> Self {
        PaEntityState {
            id: value.index as PaEntityId,
            is_muted: value.mute,
            channel_volumes: value.volume,
            metadata: PaEntityMetadata::Source {
                name: value.name.clone().unwrap_or_default().into_owned(),
                description: value.description.clone().unwrap_or_default().into_owned(),
            },
        }
    }
}

impl From<&SinkInfo<'_>> for PaEntityState {
    fn from(value: &SinkInfo) -> Self {
        PaEntityState {
            id: value.index as PaEntityId,
            is_muted: value.mute,
            channel_volumes: value.volume,
            metadata: PaEntityMetadata::Sink {
                name: value.name.clone().unwrap_or_default().into_owned(),
                description: value.description.clone().unwrap_or_default().into_owned(),
            },
        }
    }
}

impl From<&SinkInputInfo<'_>> for PaEntityState {
    fn from(value: &SinkInputInfo) -> Self {
        PaEntityState {
            id: value.index as PaEntityId,
            is_muted: value.mute,
            channel_volumes: value.volume,
            metadata: PaEntityMetadata::SinkInput {
                description: value.name.clone().unwrap_or_default().into_owned(),
                application_name: value
                    .proplist
                    .get("application.name")
                    .map(|v| String::from_utf8_lossy(v).trim_end_matches(char::from(0)).to_owned()),
                binary_name: value
                    .proplist
                    .get("application.process.binary")
                    .map(|v| String::from_utf8_lossy(v).trim_end_matches(char::from(0)).to_owned()),
            },
        }
    }
}

#[derive(Debug, Clone)]
pub struct PaVolumeState {
    timestamp: Instant,
    entities_by_id: HashMap<PaEntityId, Arc<PaEntityState>>,
    default_sink_id: PaEntityId,
    default_source_id: PaEntityId,
}

impl PaVolumeState {
    pub fn timestamp(&self) -> &Instant {
        &self.timestamp
    }

    pub fn entities_by_id(&self) -> &HashMap<PaEntityId, Arc<PaEntityState>> {
        &self.entities_by_id
    }

    pub fn default_sink_id(&self) -> &PaEntityId {
        &self.default_sink_id
    }

    pub fn default_source_id(&self) -> &PaEntityId {
        &self.default_source_id
    }
}

impl Default for PaVolumeState {
    fn default() -> Self {
        PaVolumeState {
            timestamp: Instant::now(),
            entities_by_id: HashMap::new(),
            default_sink_id: 0,
            default_source_id: 0,
        }
    }
}

#[derive(Debug)]
enum PaCommand {
    SetIsMuted { id: PaEntityId, value: bool },
    SetChannelVolumes { id: PaEntityId, channel_volumes: Box<[f32]> },
    Terminate,
}

struct PaThread {
    mainloop: Rc<RefCell<Mainloop>>,
    context: Rc<RefCell<Context>>,
    introspector: Rc<RefCell<Introspector>>,
    commands_rx: flume::Receiver<PaCommand>,
    state_tx: broadcast::Sender<Arc<PaVolumeState>>,
    current_state: Arc<RwLock<Option<Arc<PaVolumeState>>>>,
}

impl PaThread {
    fn spawn(
        client_name: String,
        commands_rx: flume::Receiver<PaCommand>,
        state_tx: broadcast::Sender<Arc<PaVolumeState>>,
        current_state: Arc<RwLock<Option<Arc<PaVolumeState>>>>,
    ) {
        thread::spawn(move || {
            let mainloop = Mainloop::new().unwrap();
            let context = Context::new(&mainloop, &client_name).unwrap();
            let introspector = context.introspect();

            let mut t = PaThread {
                mainloop: Rc::new(RefCell::new(mainloop)),
                context: Rc::new(RefCell::new(context)),
                introspector: Rc::new(RefCell::new(introspector)),
                commands_rx,
                state_tx,
                current_state,
            };

            t.init();
            t.run();
        });
    }

    fn init(&mut self) {
        self.context.borrow_mut().connect(None, FlagSet::NOFLAGS, None).unwrap();
        self.wait_for(|s| s.context.borrow().get_state() == libpulse_binding::context::State::Ready);

        let introspector = Rc::clone(&self.introspector);

        {
            let entities_by_id = Rc::new(RefCell::new(HashMap::<PaEntityId, Arc<PaEntityState>>::new()));
            let entities_by_id_c = Rc::clone(&entities_by_id);

            self.wait_for_operation(introspector.borrow().get_sink_info_list(move |list_result| match list_result {
                ListResult::Error => panic!("Introspector.get_sink_info_list failed"),
                ListResult::End => {}
                ListResult::Item(sink) => {
                    entities_by_id_c.borrow_mut().insert(sink.index as PaEntityId, Arc::new(sink.into()));
                }
            }))
            .unwrap();

            let entities_by_id_c = Rc::clone(&entities_by_id);
            self.wait_for_operation(introspector.borrow().get_source_info_list(move |list_result| match list_result {
                ListResult::Error => panic!("Introspector.get_source_info_list failed"),
                ListResult::End => {}
                ListResult::Item(source) => {
                    entities_by_id_c.borrow_mut().insert(source.index as PaEntityId, Arc::new(source.into()));
                }
            }))
            .unwrap();

            let entities_by_id_c = Rc::clone(&entities_by_id);
            self.wait_for_operation(introspector.borrow().get_sink_input_info_list(move |list_result| match list_result {
                ListResult::Error => panic!("Introspector.get_sink_input_info_list failed"),
                ListResult::End => {}
                ListResult::Item(sink_input) => {
                    let id = sink_input.index as PaEntityId;
                    entities_by_id_c.borrow_mut().insert(id, Arc::new(sink_input.into()));
                }
            }))
            .unwrap();

            PaThread::set_state(
                &self.current_state,
                &self.state_tx,
                Arc::new(PaVolumeState {
                    timestamp: Instant::now(),
                    entities_by_id: Rc::into_inner(entities_by_id).unwrap().into_inner(),
                    default_sink_id: 0,
                    default_source_id: 0,
                }),
            );
        };

        self.context
            .borrow_mut()
            .subscribe(InterestMaskSet::SINK | InterestMaskSet::SOURCE | InterestMaskSet::SINK_INPUT, |success| {
                if !success {
                    panic!("Context.subscribe failed")
                }
            });
    }

    fn run(mut self) {
        let introspector = Rc::clone(&self.introspector);
        let current_state = Arc::clone(&self.current_state);
        let state_tx = self.state_tx.clone();

        self.context
            .borrow_mut()
            .set_subscribe_callback(Some(Box::new(move |facility, operation, entity_id| {
                let entity_id = entity_id as PaEntityId;
                let facility = facility.unwrap();

                let timestamp = Instant::now();
                let current_state = Arc::clone(&current_state);

                match operation.unwrap() {
                    subscribe::Operation::Removed => {
                        let state = PaThread::unwrap_state(&current_state);

                        let entities_by_id = state.entities_by_id.without(&entity_id);
                        let default_source_id = if entity_id == state.default_source_id { 0 } else { state.default_source_id };
                        let default_sink_id = if entity_id == state.default_sink_id { 0 } else { state.default_sink_id };

                        PaThread::set_state(
                            &current_state,
                            &state_tx,
                            Arc::new(PaVolumeState {
                                timestamp,
                                entities_by_id,
                                default_source_id,
                                default_sink_id,
                            }),
                        );
                    }
                    subscribe::Operation::New | subscribe::Operation::Changed => {
                        let state_tx = state_tx.clone();

                        match facility {
                            Facility::SinkInput => {
                                introspector.borrow().get_sink_input_info(entity_id, move |list_result| match list_result {
                                    ListResult::Error => panic!("Introspector.get_sink_input_info failed"),
                                    ListResult::End => {}
                                    ListResult::Item(sink_input) => {
                                        let state = PaThread::unwrap_state(&current_state);
                                        let id = sink_input.index as PaEntityId;

                                        let entities_by_id = state.entities_by_id.update(id, Arc::new(sink_input.into()));
                                        PaThread::set_state(
                                            &current_state,
                                            &state_tx,
                                            Arc::new(PaVolumeState {
                                                timestamp,
                                                entities_by_id,
                                                ..*state
                                            }),
                                        );
                                    }
                                });
                            }
                            Facility::Sink => {
                                introspector.borrow().get_sink_info_by_index(entity_id, move |list_result| match list_result {
                                    ListResult::Error => panic!("Introspector.get_sink_info_by_index failed"),
                                    ListResult::End => {}
                                    ListResult::Item(sink_input) => {
                                        let state = PaThread::unwrap_state(&current_state);
                                        let id = sink_input.index as PaEntityId;

                                        let entities_by_id = state.entities_by_id.update(id, Arc::new(sink_input.into()));
                                        PaThread::set_state(
                                            &current_state,
                                            &state_tx,
                                            Arc::new(PaVolumeState {
                                                timestamp,
                                                entities_by_id,
                                                ..*state
                                            }),
                                        );
                                    }
                                });
                            }
                            Facility::Source => {
                                introspector.borrow().get_source_info_by_index(entity_id, move |list_result| match list_result {
                                    ListResult::Error => panic!("Introspector.get_source_info_by_index failed"),
                                    ListResult::End => {}
                                    ListResult::Item(sink_input) => {
                                        let state = PaThread::unwrap_state(&current_state);
                                        let id = sink_input.index as PaEntityId;

                                        let entities_by_id = state.entities_by_id.update(id, Arc::new(sink_input.into()));
                                        PaThread::set_state(
                                            &current_state,
                                            &state_tx,
                                            Arc::new(PaVolumeState {
                                                timestamp,
                                                entities_by_id,
                                                ..*state
                                            }),
                                        );
                                    }
                                });
                            }
                            _ => {}
                        };
                    }
                };
            })));

        let current_state = Arc::clone(&self.current_state);
        let mainloop = Rc::clone(&self.mainloop);

        loop {
            self.run_single_mainloop_iteration(false);

            while let Ok(command) = self.commands_rx.try_recv() {
                match command {
                    PaCommand::SetIsMuted { id, value } => {
                        if let Some(state) = PaThread::unwrap_state(&current_state).entities_by_id.get(&id) {
                            match state.kind() {
                                PaEntityKind::Sink => self.introspector.borrow_mut().set_sink_mute_by_index(id, value, None),
                                PaEntityKind::Source => self.introspector.borrow_mut().set_source_mute_by_index(id, value, None),
                                PaEntityKind::SinkInput => self.introspector.borrow_mut().set_sink_input_mute(id, value, None),
                            };
                        }
                    }
                    PaCommand::SetChannelVolumes { id, channel_volumes } => {
                        if let Some(state) = PaThread::unwrap_state(&current_state).entities_by_id.get(&id) {
                            let mut value = state.channel_volumes;
                            for (i, v) in channel_volumes.iter().enumerate() {
                                value.set(i as u8, Volume((Volume::NORMAL.0 as f32 * v).floor() as u32));
                            }

                            match state.kind() {
                                PaEntityKind::Sink => self.introspector.borrow_mut().set_sink_volume_by_index(id, &value, None),
                                PaEntityKind::Source => self.introspector.borrow_mut().set_source_volume_by_index(id, &value, None),
                                PaEntityKind::SinkInput => self.introspector.borrow_mut().set_sink_input_volume(id, &value, None),
                            };
                        }
                    }
                    PaCommand::Terminate => {
                        mainloop.borrow_mut().quit(Retval(0));
                    }
                }
            }
        }
    }

    fn unwrap_state(state: &Arc<RwLock<Option<Arc<PaVolumeState>>>>) -> Arc<PaVolumeState> {
        state
            .read()
            .unwrap()
            .as_ref()
            .cloned()
            .expect("this function is only called after the initial state was set")
    }

    fn set_state(current_state: &RwLock<Option<Arc<PaVolumeState>>>, state_tx: &broadcast::Sender<Arc<PaVolumeState>>, value: Arc<PaVolumeState>) {
        let mut s = current_state.write().unwrap();
        *s = Some(Arc::clone(&value));

        // If there are no subscribers, that’s ok.
        _ = state_tx.send(value);
    }

    fn run_single_mainloop_iteration(&mut self, block: bool) {
        match self.mainloop.borrow_mut().iterate(block) {
            IterateResult::Quit(_) => {
                panic!("Mainloop quit.")
            }
            IterateResult::Err(e) => {
                panic!("Mainloop error: {}", e)
            }
            IterateResult::Success(_) => {}
        }
    }

    fn wait_for(&mut self, predicate: impl Fn(&mut Self) -> bool) {
        loop {
            self.run_single_mainloop_iteration(true);

            if predicate(self) {
                break;
            }
        }
    }

    fn wait_for_operation<T: ?Sized>(&mut self, operation: Operation<T>) -> Result<(), ()> {
        loop {
            self.run_single_mainloop_iteration(true);

            match operation.get_state() {
                libpulse_binding::operation::State::Running => {}
                libpulse_binding::operation::State::Done => return Ok(()),
                libpulse_binding::operation::State::Cancelled => return Err(()),
            }
        }
    }
}

#[derive(Debug)]
struct PaWorker {
    commands_tx: flume::Sender<PaCommand>,
}

impl Drop for PaWorker {
    fn drop(&mut self) {
        self.commands_tx.send(PaCommand::Terminate).ok();
    }
}

#[derive(Debug, Clone)]
pub struct PaVolumeInterface {
    #[allow(unused)]
    worker: Arc<PaWorker>,
    current_state: Arc<RwLock<Option<Arc<PaVolumeState>>>>,
    state_tx: broadcast::Sender<Arc<PaVolumeState>>,
    commands_tx: flume::Sender<PaCommand>,
}

impl PaVolumeInterface {
    pub fn spawn_thread(client_name: String) -> PaVolumeInterface {
        let (commands_tx, commands_rx) = flume::unbounded();
        let state_tx = broadcast::Sender::new(5);
        let current_state = Arc::new(RwLock::new(None));

        PaThread::spawn(client_name, commands_rx, state_tx.clone(), Arc::clone(&current_state));

        let worker = PaWorker {
            commands_tx: commands_tx.clone(),
        };

        PaVolumeInterface {
            worker: Arc::new(worker),
            current_state,
            commands_tx,
            state_tx,
        }
    }

    pub fn subscribe_to_state(&self) -> (Option<Arc<PaVolumeState>>, Receiver<Arc<PaVolumeState>>) {
        let rx = self.state_tx.subscribe();
        let state = self.current_state();
        (state, rx)
    }

    pub fn current_state(&self) -> Option<Arc<PaVolumeState>> {
        self.current_state.read().unwrap().clone()
    }

    pub fn set_is_muted(&self, id: PaEntityId, value: bool) {
        self.commands_tx.send(PaCommand::SetIsMuted { id, value }).unwrap()
    }

    pub fn set_channel_volumes(&self, id: PaEntityId, channel_volumes: impl Into<Box<[f32]>>) {
        self.commands_tx
            .send(PaCommand::SetChannelVolumes {
                id,
                channel_volumes: channel_volumes.into(),
            })
            .unwrap()
    }
}
