use std::collections::HashMap;
use std::ffi::OsString;
use std::path::Path;
use std::process::Stdio;
use std::sync::Arc;

use color_eyre::eyre::{eyre, WrapErr};
use color_eyre::Result;
use is_executable::IsExecutable;
use tokio::io::{AsyncBufReadExt, AsyncWriteExt, BufReader};
use tokio::process::{ChildStdin, Command};

use deckster_shared::handler_communication::{
    HandlerCommand, HandlerEvent, HandlerInitializationError, HandlerInitializationResultMessage, InitialHandlerMessage,
};
use deckster_shared::path::{KeyPath, KnobPath};

pub struct KeyOrKnobConfig {
    pub handler_name: Box<str>,
    pub handler_config: Arc<toml::Table>,
}

pub async fn start(
    handlers_directory: &Path,
    key_configs: HashMap<KeyPath, KeyOrKnobConfig>,
    knob_configs: HashMap<KnobPath, KeyOrKnobConfig>,
    commands_sender: flume::Sender<HandlerCommand>,
    mut events_receiver: tokio::sync::broadcast::Receiver<HandlerEvent>,
) -> Result<()> {
    let handler_names: Vec<OsString> = std::fs::read_dir(handlers_directory)
        .wrap_err_with(|| format!("while reading the handlers directory: {}", handlers_directory.to_string_lossy()))?
        .filter_map(|entry| {
            if let Ok(entry) = entry {
                let path = entry.path();
                if path.is_executable() {
                    return Some(path.file_name().unwrap().to_os_string());
                }
            }

            None
        })
        .collect();

    let mut handler_stdin_by_name: HashMap<Box<str>, ChildStdin> = HashMap::with_capacity(handler_names.len());

    for handler_name in handler_names {
        let handler_name = handler_name.into_string().map_err(|_| eyre!("Command names must be valid Unicode."))?;

        let mut command = Command::new(handlers_directory.join(&handler_name))
            .arg("deckster-run")
            .stdin(Stdio::piped())
            .stdout(Stdio::piped())
            .spawn()
            .wrap_err_with(|| format!("while spawning handler: {}", handler_name))?;

        let mut stdout_lines = BufReader::new(command.stdout.take().unwrap()).lines();
        let mut stdin = command.stdin.unwrap();

        let initial_handler_message = InitialHandlerMessage {
            key_configs: key_configs
                .iter()
                .filter_map(|(path, c)| {
                    if *c.handler_name == handler_name {
                        Some((path.clone(), Arc::clone(&c.handler_config)))
                    } else {
                        None
                    }
                })
                .collect(),
            knob_configs: knob_configs
                .iter()
                .filter_map(|(path, c)| {
                    if *c.handler_name == handler_name {
                        Some((path.clone(), Arc::clone(&c.handler_config)))
                    } else {
                        None
                    }
                })
                .collect(),
        };

        let serialized_message = serde_json::to_string(&initial_handler_message).unwrap().into_boxed_str().into_boxed_bytes();

        stdin.write_all(&serialized_message).await.unwrap();
        stdin.write_u8(b'\n').await.unwrap();
        stdin.flush().await.unwrap();

        let result_line = stdout_lines.next_line().await?.unwrap();
        let result: HandlerInitializationResultMessage = serde_json::from_str(&result_line)?;

        if let HandlerInitializationResultMessage::Error { error } = result {
            #[rustfmt::skip]
            if let HandlerInitializationError::InvalidConfig { supports_keys, supports_knobs, .. } = error {
                if !supports_keys && !initial_handler_message.key_configs.is_empty() {
                    return Err(eyre!(
                        "The '{handler_name}' handler does not support keys, but these keys tried to use it: {:?}",
                        initial_handler_message.key_configs.keys().map(|k| k.to_string()).collect::<String>()
                    ));
                } else if !supports_knobs && !initial_handler_message.knob_configs.is_empty() {
                    return Err(eyre!(
                        "The '{handler_name}' handler does not support knobs, but these knobs tried to use it: {:?}",
                        initial_handler_message.knob_configs.keys().map(|k| k.to_string()).collect::<String>()
                    ));
                }
            };

            return Err(eyre!("Starting the '{handler_name}' handler failed: {error}"));
        }

        let commands_sender = commands_sender.clone();

        tokio::spawn(async move {
            while let Ok(Some(line)) = stdout_lines.next_line().await {
                if line.starts_with('{') {
                    let command = serde_json::from_str::<HandlerCommand>(&line).unwrap();

                    commands_sender.send_async(command).await.unwrap();
                } else {
                    println!("{}", line);
                }
            }
        });

        handler_stdin_by_name.insert(handler_name.into_boxed_str(), stdin);
    }

    tokio::spawn(async move {
        while let Ok(event) = events_receiver.recv().await {
            let config = match &event {
                HandlerEvent::Key { path, .. } => {
                    if let Some(config) = key_configs.get(path) {
                        config
                    } else {
                        continue;
                    }
                }
                HandlerEvent::Knob { path, .. } => {
                    if let Some(config) = knob_configs.get(path) {
                        config
                    } else {
                        continue;
                    }
                }
            };

            let handler_stdin = handler_stdin_by_name.get_mut(&config.handler_name).expect("was already checked above");
            let serialized_event = serde_json::to_string(&event).unwrap().into_boxed_str().into_boxed_bytes();

            handler_stdin.write_all(&serialized_event).await.unwrap();
            handler_stdin.write_u8(b'\n').await.unwrap();
            handler_stdin.flush().await.unwrap();
        }
    });

    Ok(())
}
