use std::str::FromStr;
use std::time::Duration;

use rumqttc::{Event, Incoming, MqttOptions, QoS};
use tokio::sync::broadcast;

use deckster_shared::handler_communication::{HandlerCommand, HandlerEvent};
use deckster_shared::path::{KeyPath, KeyPosition, KnobPath, KnobPosition};
use deckster_shared::style::{KeyStyle, KnobStyle};

use crate::model::config::MqttConfig;

pub async fn start_mqtt_client(config: &MqttConfig, commands_sender: flume::Sender<HandlerCommand>, mut events_receiver: broadcast::Receiver<HandlerEvent>) {
    let topic_prefix = config.topic_prefix.to_owned();

    let mut options = MqttOptions::new(&config.topic_prefix, &config.host, config.port);
    options.set_keep_alive(Duration::from_secs(3));
    options.set_clean_session(false);

    if let Some(credentials) = &config.credentials {
        options.set_credentials(&credentials.username, &credentials.password);
    }

    let (client, mut event_loop) = rumqttc::AsyncClient::new(options, 32);

    client.subscribe(format!("{topic_prefix}/keys/+/+/style"), QoS::ExactlyOnce).await.unwrap();
    client.subscribe(format!("{topic_prefix}/knobs/+/+/style"), QoS::ExactlyOnce).await.unwrap();
    client.subscribe(format!("{topic_prefix}/knobs/+/+/value"), QoS::ExactlyOnce).await.unwrap();

    tokio::spawn({
        let topic_prefix = topic_prefix.clone();

        async move {
            while let Ok(event) = events_receiver.recv().await {
                match event {
                    HandlerEvent::Key { path, event } => {
                        client
                            .publish(
                                format!("{topic_prefix}/keys/{path}/events"),
                                QoS::ExactlyOnce,
                                false,
                                serde_json::to_string(&event).unwrap(),
                            )
                            .await
                            .unwrap();
                    }
                    HandlerEvent::Knob { path, event } => {
                        client
                            .publish(
                                format!("{topic_prefix}/knobs/{path}/events"),
                                QoS::ExactlyOnce,
                                false,
                                serde_json::to_string(&event).unwrap(),
                            )
                            .await
                            .unwrap();
                    }
                }
            }
        }
    });

    tokio::spawn(async move {
        while let Ok(event) = event_loop.poll().await {
            if let Event::Incoming(Incoming::Publish(event)) = event {
                let segments = event.topic.strip_prefix(&topic_prefix).unwrap().split('/').collect::<Vec<&str>>();
                let page_id = segments[1];
                let position = segments[2];
                let property = segments[3];

                match segments[0] {
                    "keys" => {
                        if property == "style" {
                            let value = serde_json::from_slice::<Option<KeyStyle>>(&event.payload).unwrap();

                            commands_sender
                                .send(HandlerCommand::SetKeyStyle {
                                    path: KeyPath {
                                        page_id: page_id.to_owned(),
                                        position: KeyPosition::from_str(position).unwrap(),
                                    },
                                    value,
                                })
                                .unwrap();
                        }
                    }
                    "knobs" => {
                        let position = KnobPosition::from_str(position).unwrap();

                        match property {
                            "style" => {
                                let value = serde_json::from_slice::<Option<KnobStyle>>(&event.payload).unwrap();

                                commands_sender
                                    .send(HandlerCommand::SetKnobStyle {
                                        path: KnobPath {
                                            page_id: page_id.to_owned(),
                                            position,
                                        },
                                        value,
                                    })
                                    .unwrap();
                            }
                            "value" => {
                                let value = serde_json::from_slice::<Option<f32>>(&event.payload).unwrap();

                                commands_sender
                                    .send(HandlerCommand::SetKnobValue {
                                        path: KnobPath {
                                            page_id: page_id.to_owned(),
                                            position,
                                        },
                                        value,
                                    })
                                    .unwrap();
                            }
                            _ => {}
                        }
                    }
                    _ => {}
                };
            }
        }
    });
}
